\ SOL Turtle Stepper Driver
\ By Nils M Holm, 2004

: robot.scr robot.scr ;  dict>

\ Read Programmable Interrupt Timer (PIT)
: pit@ ( -- n )  [hex] 40 cp@ [hex] 40 cp@  8 shl or ;

\ Re-set count-down timer roughly to 0
: 0pit! ( -- )  10000 for  10000 for  pit@  60000 u> if  unloop
      exit then  next  next  abort" timer reset" ;

\ Measure loop time: iterations --> ticks
: looptime ( i -- t )  nointr  [hex] 06 [hex] 43 cp!  intr
  0pit!  pit@  swap for next  pit@  - abs ;

\ Make sure there's a working PIT
: ?pit ( -- )  100 looptime 1+  dup 2 *  swap 4 *
  300 looptime 1+  -rot  within not if  abort" broken timer"
    then ;

\ Find an upper limit for the iteration count
: maxloop ( -- n )  1  15 for  dup looptime
    2168 > if  rdrop exit then  2 *  next
  abort" computer too fast" ;

\ Find the proper iteration count to waste a given amount
\ of time; using binary search
: adjust ( -- n )  ?pit  maxloop  dup 2 /  swap  begin
    dup looptime  dup 2168 <  if  drop over +  else
    drop over -  then  swap 2 / swap  over  0= until  nip ;

adjust var ud   \ Micro delay counter
    10 var d    \ Delay between stepper pulses

\ Sleep for a the length of a micro delay
: w ( -- ) ud @ for next ;

\ Wait for N micro delays
: w ( n -- ) for w next ;

\ Create sequence header
: seq ( -- )  create  link ;

\ Create sequence of eight patterns
: :seq ( -- ) seq  base @  hex!  8 for  #in c,  next
  base ! ;

\ Combine two sequences
: combine ( a1 a2 -- )  8 for  2dup  c@ swap c@  or  c,
    1+ swap 1+  next  2drop ;

\ Stepper sequences (using half stepping)
:seq lf  09 08 0c 04 06 02 03 01   \ left motor forward
:seq lb  01 03 02 06 04 0c 08 09   \ left motor backward
:seq rf  10 30 20 60 40 c0 80 90   \ right motor forward
:seq rb  90 80 c0 40 60 20 30 10   \ right motor backward
 seq fw  lf rf combine             \ drive forward
 seq bw  lb rb combine             \ drive backward
 seq tl  lb rf combine             \ turn left
 seq tr  lf rb combine             \ turn right

\ Wait, then activate some coils of the steppers
: s ( c -- ) d @ w  [hex] 278 cp! ;

\ Check sensor
: sensor@ ( pos -- f ) [hex] 279 cp@  swap shr  1 and  0= ;

\ Check left sensor
: ls@ ( -- f ) 3 sensor@ ;

\ Check right sensor
: rs@ ( -- f ) 4 sensor@ ;

\ Turtle blocked?
: blocked ( -- f )  ls@ rs@ or ;

\ Stop motors
: stop ( -- ) 0 s ;

\ Send stepper sequence to motor
: step ( a -- )  8 for  dup c@  s  1+  next  drop ;

\ Slow down motor to half speed, return old speed
: slow ( -- n )  d @  dup  2 *  d ! ;

\ Restore original motor speed
: old ( n -- )  d ! ;

\ Turn to left/right
: turn ( n a -- n )  slow >r  swap  for  dup step  blocked
    if  drop  r>  r> old  exit  then  next  drop  r> old  0 ;

\ Abort, if Turtle blocked
: ?blocked ( f -- )  if  abort" turtle blocked"  then ;

\ Turn left N units
: left ( n -- )  tl turn  ?blocked ;

\ Turn right N units
: right ( n -- )  tr turn  ?blocked ;

\ Accelerate
: accel ( -- )  7 for  fw step  -1 d +!  next ;

\ Brake
: brake ( -- )  7 for  fw step   1 d +!  next ;

\ Drive the turtle
: drive ( n a -- n )  swap  for  dup step  blocked if  drop
      r>  exit  then  next  drop  0 ;

\ Drive forward N units; drive fast if N>20
: forward ( n -- )  dup 20 > if  accel  14 -  fw drive  brake
    else  fw drive  then  ?blocked ;

\ Drive backward N units
: backward ( n -- ) bw drive  ?blocked ;

<dict robot  robot show
